import numpy as np

import tensortrax as tr
import tensortrax.math as tm


def simple(F):
    return tm.trace(F)


def neo_hooke(F):
    C = F.T @ F
    I1 = tm.trace(C)
    J = tm.linalg.det(F)
    return (J ** (-2 / 3) * I1 - 3) / 2


def ogden(F, mu=1, alpha=2):
    C = F.T @ F
    J = tm.linalg.det(F)
    λ = tm.sqrt(tm.linalg.eigvalsh(J ** (-2 / 3) * C))
    return tm.sum(1 / alpha * (λ**alpha - 1))


def test_hvp():
    F = δF = ΔF = np.tile(
        (np.eye(3).ravel() + np.arange(9) / 10).reshape(3, 3, 1, 1), (1, 1, 2, 4)
    )

    for parallel in [False, True]:
        for fun in [simple, neo_hooke, ogden]:
            δfun = tr.gradient_vector_product(fun, wrt="F", ntrax=2, parallel=parallel)(
                F=F, δx=δF
            )
            Δδfun = tr.hessian_vectors_product(
                fun, wrt="F", ntrax=2, parallel=parallel
            )(F=F, δx=δF, Δx=ΔF)

            hvp = tr.hessian_vector_product(fun, wrt="F", ntrax=2, parallel=parallel)(
                F=F, δx=δF
            )

            assert δfun.shape == (2, 4)
            assert Δδfun.shape == (2, 4)

            assert hvp.shape == (3, 3, 2, 4)

            assert not np.any(np.isnan(δfun))
            assert not np.any(np.isnan(Δδfun))
            assert not np.any(np.isnan(hvp))


def test_gvp():
    F = np.diag([2, 1.15, 1.15])
    F[:, 0] += np.array([0, -0.15, -0.15])
    δF = np.vstack([np.zeros(3), np.zeros(3), np.array([-0.04, 0.04, 0.04])])

    δψ = tr.gradient_vector_product(neo_hooke)(F, δx=δF)
    δψ_reference = tm.special.ddot(tr.gradient(neo_hooke)(F), δF)

    assert np.isclose(δψ, δψ_reference)


if __name__ == "__main__":
    test_hvp()
    test_gvp()



"""
tensorTRAX: Math on (Hyper-Dual) Tensors with Trailing Axes.
"""

from copy import copy
from functools import wraps

import numpy as np
from joblib import Parallel, cpu_count, delayed

from ._tensor import Tensor, Δδ, broadcast_to, f, δ
from .math.special import from_triu_1d, from_triu_2d, triu_1d


def take(fun, item=0):
    "Evaluate the function and take only the selected item."

    @wraps(fun)
    def evaluate(*args, **kwargs):
        return fun(*args, **kwargs)[item]

    return evaluate


def add_tensor(
    args, kwargs, wrt, ntrax, sym=False, gradient=False, hessian=False, δx=None, Δx=None
):
    "Modify the arguments and replace the w.r.t.-argument by a tensor."

    kwargs_out = copy(kwargs)
    args_out = list(args)

    if isinstance(wrt, str):
        args_old = kwargs
        args_new = kwargs_out
    elif isinstance(wrt, int):
        args_old = args
        args_new = args_out
    else:
        raise TypeError(
            f"Type of wrt not supported. type(wrt) is {type(wrt)} (must be str or int)."
        )

    x = args_old[wrt]

    if sym:
        x = triu_1d(x)

    tensor = Tensor(x=x, ntrax=ntrax)
    trax = tensor.trax

    tensor.init(gradient=gradient, hessian=hessian, sym=sym, δx=δx, Δx=Δx)

    if sym:
        tensor = from_triu_1d(tensor)

    args_new[wrt] = tensor

    return args_out, kwargs_out, tensor.shape, trax


def partition(args, kwargs, wrt, ntrax, parallel, chunks=None, batch=100, axis=None):
    """Partition function (keyword) arguments into a list of (keyword) arguments. Only
    top-level args and kwargs with equal shapes to be splitted are allowed."""

    # deactivate parallel evaluation if no trailing axes are present
    if ntrax == 0:
        parallel = False

    # get shape of trailing axes, define axis and chunks
    # if size of chosen axis is below batch, deactivate parallel evaluation
    if parallel:
        # get shape of trailing axes
        trax = (kwargs[wrt] if isinstance(wrt, str) else args[wrt]).shape[-ntrax:]

        # select axis
        if axis is None:
            axis = -(1 + np.argmax(trax[::-1]))

        # define chunks
        if chunks is None:
            if (trax[axis] // batch) > 0:
                chunks = min(trax[axis] // batch, cpu_count())
            else:
                parallel = False

    if not parallel:
        list_of_args_kwargs = [(args, kwargs)]
        chunks = 1
        axis = -1

    else:
        # generate list with args and kwargs for chunks
        list_of_args_kwargs = [[list(args), {**kwargs}] for chunk in range(chunks)]

        # test if object has attribute shape (is tensor or array)
        def isactive(x):
            return hasattr(x, "shape") and np.all(np.isin(trax, x.shape[-ntrax:]))

        # iterate through args and split tensor-like objects
        args_partitioned = []
        for i, arg in enumerate(args):
            if isactive(arg):
                args_partitioned.append((i, np.array_split(arg, chunks, axis=axis)))

        # replace arguments by chunks
        for i, arg in enumerate(list_of_args_kwargs):
            for j, argp in args_partitioned:
                list_of_args_kwargs[i][0][j] = argp[i]

        # iterate through kwargs and split tensor-like objects
        kwargs_partitioned = []
        for key, value in kwargs.items():
            if isactive(value):
                kwargs_partitioned.append(
                    (key, np.array_split(value, chunks, axis=axis))
                )

        # replace keyword arguments by chunks
        for i, kwarg in enumerate(list_of_args_kwargs):
            for key, value in kwargs_partitioned:
                list_of_args_kwargs[i][1][key] = value[i]

    return list_of_args_kwargs, chunks, axis


def concatenate_results(res, axis, full_output):
    "Concatenate results (with optional full output) on existing axis."

    def concat(arrays, axis):
        "Concatenate arrays, fall-back to first item if shape of first array is zero."

        if len(arrays[0].shape) == 0 or len(arrays) == 1:
            return arrays[0]
        else:
            return np.concatenate(arrays, axis=axis)

    if full_output:
        nres = len(res[0])
        return [concat([r[a] for r in res], axis=axis) for a in range(nres)]
    else:
        return concat(res, axis=axis)


def function(fun, wrt=0, ntrax=0, parallel=False):
    r"""Evaluate a function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0).
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the function in parallel (threaded).

    Returns
    -------
    ndarray
        NumPy array containing the function result.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>>
    >>> F.shape
    (3, 3, 8, 20)

    >>> W = tr.function(fun, wrt=0, ntrax=2)(F)
    >>> W = tr.function(fun, wrt="F", ntrax=2)(F=F)
    >>>
    >>> W.shape
    >>> (8, 20)
    """

    @wraps(fun)
    def evaluate_function(*args, **kwargs):
        def kernel(args, kwargs):
            args, kwargs, shape, trax = add_tensor(
                args, kwargs, wrt, ntrax, False, False, False
            )
            func = fun(*args, **kwargs)
            if isinstance(func, Tensor):
                func = f(func)
            return func

        list_of_args_kwargs, chunks, axis = partition(
            args, kwargs, wrt, ntrax, parallel
        )

        res = Parallel(n_jobs=chunks, prefer="threads")(
            delayed(kernel)(*args_chunk) for args_chunk in list_of_args_kwargs
        )

        return concatenate_results(res=res, axis=axis, full_output=False)

    return evaluate_function


def gradient(fun, wrt=0, ntrax=0, parallel=False, full_output=False, sym=False):
    r"""Evaluate the gradient of a scalar-valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The gradient is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the gradient in parallel (threaded).
    full_output: bool, optional
        Return the gradient and the function (default is False).
    sym : bool, optional
        Apply the variations only on the upper-triangle entries of a symmetric second
        order tensor. This is a performance feature and requires no modification of the
        callable ``fun`` and the input arguments, including ``wrt``. Default is False.

    Returns
    -------
    ndarray or list of ndarray
        NumPy array containing the gradient result. If ``full_output=True``, the
        function is also returned.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>>
    >>> F.shape
    (3, 3, 8, 20)

    >>> dWdF = tr.gradient(fun, wrt=0, ntrax=2)(F)
    >>> dWdF = tr.gradient(fun, wrt="F", ntrax=2)(F=F)
    >>>
    >>> dWdF.shape
    >>> (3, 3, 8, 20)
    """

    @wraps(fun)
    def evaluate_gradient(*args, **kwargs):
        def kernel(args, kwargs):
            args, kwargs, shape, trax = add_tensor(
                args, kwargs, wrt, ntrax, sym, True, False
            )
            func = fun(*args, **kwargs)
            grad = δ(func) if sym is False else from_triu_1d(δ(func))
            grad = broadcast_to(grad, (*shape, *trax))
            if full_output:
                trax = (1,) if len(trax) == 0 else trax
                zeros = np.zeros_like(shape) if sym is False else (0,)
                funct = f(func)[(*zeros,)]
                funct = broadcast_to(funct, (*trax,))
                return grad, funct
            else:
                return grad

        list_of_args, chunks, axis = partition(args, kwargs, wrt, ntrax, parallel)

        res = Parallel(n_jobs=chunks, prefer="threads")(
            delayed(kernel)(*args_chunk) for args_chunk in list_of_args
        )

        return concatenate_results(res=res, axis=axis, full_output=full_output)

    return evaluate_gradient


def hessian(fun, wrt=0, ntrax=0, parallel=False, full_output=False, sym=False):
    r"""Evaluate the Hessian of a scalar-valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The Hessian is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the Hessian in parallel (threaded).
    full_output: bool, optional
        Return the hessian, the gradient and the function (default is False).
    sym : bool, optional
        Apply the variations only on the upper-triangle entries of a symmetric second
        order tensor. This is a performance feature and requires no modification of the
        callable ``fun`` and the input arguments, including ``wrt``. Default is False.

    Returns
    -------
    ndarray or list of ndarray
        NumPy array containing the Hessian result. If ``full_output=True``, the
        gradient and the function are also returned.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>>
    >>> F.shape
    (3, 3, 8, 20)

    >>> d2WdFdF = tr.hessian(fun, wrt=0, ntrax=2)(F)
    >>> d2WdFdF = tr.hessian(fun, wrt="F", ntrax=2)(F=F)
    >>>
    >>> d2WdFdF.shape
    >>> (3, 3, 3, 3, 8, 20)
    """

    @wraps(fun)
    def evaluate_hessian(*args, **kwargs):
        def kernel(args, kwargs):
            args, kwargs, shape, trax = add_tensor(
                args, kwargs, wrt, ntrax, sym, False, True
            )
            func = fun(*args, **kwargs)
            hess = Δδ(func) if sym is False else from_triu_2d(Δδ(func))

            if full_output:
                grad = δ(func) if sym is False else from_triu_1d(δ(func))
                zeros = np.zeros_like(shape) if sym is False else (0,)
                grad = grad[(*[slice(None) for a in shape], *zeros)]
                grad = broadcast_to(grad, (*shape, *trax))
                funct = f(func)[
                    (
                        *zeros,
                        *zeros,
                    )
                ]
                funct = broadcast_to(funct, (*trax,))
                trax = (1,) if len(trax) == 0 else trax
                return hess, grad, funct
            else:
                return hess

        list_of_args, chunks, axis = partition(args, kwargs, wrt, ntrax, parallel)

        res = Parallel(n_jobs=chunks, prefer="threads")(
            delayed(kernel)(*args_chunk) for args_chunk in list_of_args
        )

        return concatenate_results(res=res, axis=axis, full_output=full_output)

    return evaluate_hessian


def jacobian(fun, wrt=0, ntrax=0, parallel=False, full_output=False):
    r"""Evaluate the Jacobian of a tensor-valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The Jacobian is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the Jacobian in parallel (threaded).
    full_output: bool, optional
        Return the Jacobian and the function (default is False).

    Returns
    -------
    ndarray
        NumPy array containing the Jacobian result.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(C, mu=1):
    >>>     I3 = tm.linalg.det(C)
    >>>     return mu * tm.special.dev(I3 ** (-1 / 3) * C) @ tm.linalg.inv(C)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>> C = np.einsum("ki...,kj...->ij...", F, F)
    >>>
    >>> C.shape
    (3, 3, 8, 20)

    >>> dSdC = tr.jacobian(fun, wrt=0, ntrax=2)(C)
    >>> dSdC = tr.jacobian(fun, wrt="C", ntrax=2)(C=C)
    >>>
    >>> dSdC.shape
    >>> (3, 3, 3, 3, 8, 20)
    """

    @wraps(fun)
    def evaluate_jacobian(*args, **kwargs):
        def kernel(args, kwargs):
            args, kwargs, shape, trax = add_tensor(
                args, kwargs, wrt, ntrax, False, True, False
            )
            func = fun(*args, **kwargs)

            if full_output:
                return δ(func), f(func).reshape(*func.shape, *trax)
            else:
                return δ(func)

        list_of_args, chunks, axis = partition(args, kwargs, wrt, ntrax, parallel)

        res = Parallel(n_jobs=chunks, prefer="threads")(
            delayed(kernel)(*args_chunk) for args_chunk in list_of_args
        )

        return concatenate_results(res=res, axis=axis, full_output=full_output)

    return evaluate_jacobian


def gradient_vector_product(fun, wrt=0, ntrax=0, parallel=False):
    r"""Evaluate the gradient-vector-product of a scalar-valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated. Its signature is extended to
        :func:`fun(*args, δx, **kwargs)`, where the added ``δx``-argument is the vector
        of the gradient-vector product.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The gradient-vector-product is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the gradient-vector-product in parallel (threaded).

    Returns
    -------
    ndarray
        NumPy array containing the gradient-vector-product result.

    Notes
    -----
    The *vector* :math:`\delta x` and the tensor-argument ``wrt`` must have equal or
    broadcast-compatible shapes. This means that the *vector* is not restricted to be a
    one-dimensional array but must be an array with compatible shape instead.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>> F.shape
    (3, 3, 8, 20)

    >>> np.random.seed(63254)
    >>> δF = np.random.rand(3, 3, 8, 20) / 10
    >>> δF.shape
    (3, 3, 8, 20)

    >>> dW = tr.gradient_vector_product(fun, wrt=0, ntrax=2)(F, δx=δF)
    >>> dW.shape
    >>> (8, 20)
    """

    @wraps(fun)
    def evaluate_gradient_vector_product(*args, δx, **kwargs):
        args, kwargs, shape, trax = add_tensor(
            args, kwargs, wrt, ntrax, False, gradient=True, δx=δx
        )
        δfun = δ(fun(*args, **kwargs))
        trim = np.zeros(len(δfun.shape) - ntrax, dtype=int)
        return broadcast_to(δfun[(*trim,)], trax)

    return evaluate_gradient_vector_product


def hessian_vector_product(fun, wrt=0, ntrax=0, parallel=False):
    r"""Evaluate the Hessian-vector-product of a scalar-valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated. Its signature is extended to
        :func:`fun(*args, δx, **kwargs)`, where the added ``δx``-argument is the vector
        of the Hessian-vector product.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The Hessian-vector-product is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the gradient-vector-product in parallel (threaded).

    Returns
    -------
    ndarray
        NumPy array containing the Hessian-vector-product result.

    Notes
    -----
    The *vector* :math:`\delta x` and the tensor-argument ``wrt`` must have equal or
    broadcast-compatible shapes. This means that the *vector* is not restricted to be a
    one-dimensional array but must be an array with compatible shape instead.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>> F.shape
    (3, 3, 8, 20)

    >>> np.random.seed(63254)
    >>> δF = np.random.rand(3, 3, 8, 20) / 10
    >>> δF.shape
    (3, 3, 8, 20)

    >>> dP = tr.hessian_vector_product(fun, wrt=0, ntrax=2)(F, δx=δF)
    >>> dP.shape
    >>> (3, 3, 8, 20)
    """

    @wraps(fun)
    def evaluate_hessian_vector_product(*args, δx, **kwargs):
        args, kwargs, shape, trax = add_tensor(
            args, kwargs, wrt, ntrax, False, hessian=True, δx=δx
        )
        Δδfun = Δδ(fun(*args, **kwargs))
        trim = np.zeros(len(Δδfun.shape) - len(shape) - ntrax, dtype=int)
        return broadcast_to(Δδfun[(*trim,)], (*shape, *trax))

    return evaluate_hessian_vector_product


def hessian_vectors_product(fun, wrt=0, ntrax=0, parallel=False):
    r"""Evaluate the vector-Hessian-vector- or Hessian-vectors-product of a scalar-
    valued function.

    Parameters
    ----------
    fun : callable
        The function to be evaluated. Its signature is extended to
        :func:`fun(*args, δx, Δx, **kwargs)`, where the added ``δx``- and ``Δx``-
        arguments are the vectors of the Hessian-vectors product.
    wrt : int or str, optional
        The input argument which will be treated as :class:`~tensortrax.Tensor` (default
        is 0). The Hessian-vectors-product is carried out with respect to this argument.
    ntrax : int, optional
        Number of elementwise-operating trailing axes (batch dimensions). Default is 0.
    parallel : bool, optional
        Flag to evaluate the gradient-vector-product in parallel (threaded).

    Returns
    -------
    ndarray
        NumPy array containing the Hessian-vectors-product result.

    Notes
    -----
    The *vectors* :math:`\delta x` and :math:`\Delta x` as well as the tensor-argument
    ``wrt`` must have equal or broadcast-compatible shapes. This means that the
    *vectors* are not restricted to be one-dimensional arrays but must be arrays with
    compatible shapes instead.

    Examples
    --------

    >>> import numpy as np
    >>> import tensortrax as tr
    >>> import tensortrax.math as tm
    >>>
    >>> def fun(F, mu=1):
    >>>     C = F.T @ F
    >>>     I1 = tm.trace(C)
    >>>     J = tm.linalg.det(F)
    >>>     return mu / 2 * (J ** (-2 / 3) * I1 - 3)
    >>>
    >>> np.random.seed(125161)
    >>> F = (np.eye(3) + np.random.rand(20, 8, 3, 3) / 10).T
    >>> F.shape
    (3, 3, 8, 20)

    >>> np.random.seed(63254)
    >>> δF = np.random.rand(3, 3, 8, 20) / 10
    >>> δF.shape
    (3, 3, 8, 20)

    >>> np.random.seed(85476)
    >>> ΔF = np.random.rand(3, 3, 8, 20) / 10
    >>> ΔF.shape
    (3, 3, 8, 20)

    >>> ΔδW = tr.hessian_vectors_product(fun, wrt=0, ntrax=2)(F, δx=δF, Δx=ΔF)
    >>> ΔδW.shape
    >>> (8, 20)
    """

    @wraps(fun)
    def evaluate_hessian_vectors_product(*args, δx, Δx, **kwargs):
        args, kwargs, shape, trax = add_tensor(
            args, kwargs, wrt, ntrax, False, hessian=True, δx=δx, Δx=Δx
        )
        Δδfun = Δδ(fun(*args, **kwargs))
        trim = np.zeros(len(Δδfun.shape) - ntrax, dtype=int)
        return broadcast_to(Δδfun[(*trim,)], trax)

    return evaluate_hessian_vectors_product



"""
tensorTRAX: Math on (Hyper-Dual) Tensors with Trailing Axes.
"""
import numpy as np

from .._tensor import Tensor, f


def eye(A):
    "Identity (Eye) of a Tensor."
    if isinstance(A, Tensor):
        B = np.zeros_like(f(A))
    else:
        B = np.zeros_like(A)
    B[np.diag_indices(B.shape[0])] = 1
    return B


def cross(a, b):
    "Cross product of two vectors a and b."
    return np.einsum(
        "...i->i...", np.cross(np.einsum("i...->...i", a), np.einsum("i...->...i", b))
    )



"""
tensorTRAX: Math on (Hyper-Dual) Tensors with Trailing Axes.
"""


def f(A):
    "Return the value of the tensor."
    return A.x


def δ(A):
    "Return the (dual) δ-variation of the tensor."
    return A.δx


def Δ(A):
    "Return the (dual) Δ-variation of the tensor."
    return A.Δx


def Δδ(A):
    "Return the (dual) Δδ-variation of the tensor."
    return A.Δδx



from ._special_tensor import (
    ddot,
    dev,
    erf,
    from_triu_1d,
    from_triu_2d,
    sym,
    tresca,
    triu_1d,
    try_stack,
    von_mises,
)

__all__ = [
    "ddot",
    "dev",
    "erf",
    "from_triu_1d",
    "from_triu_2d",
    "sym",
    "tresca",
    "triu_1d",
    "try_stack",
    "von_mises",
]



"""
tensorTRAX: Math on (Hyper-Dual) Tensors with Trailing Axes.
"""

from copy import deepcopy

import numpy as np

from ._helpers import Δ, Δδ, f, δ


class Tensor:
    r"""A Hyper-Dual Tensor.

    Parameters
    ----------
    x : array_like
        The data of the tensor.
    δx : array_like or None, optional
        (Dual) variation data (δ-operator) of the tensor.
    Δx : array_like or None, optional
        (Dual) variation data (Δ-operator) of the tensor.
    Δδx : array_like or None, optional
        (Dual) linearization data (Δδ-operator) of the tensor.
    ntrax : int, optional
        Total number of trailing axes including dual axes (default is 0).
    ndual : int, optional
        Number of axes containing dual data (default is 0).

    Examples
    --------
    A hyper-dual tensor consists of batches of tensors, where the elementwise-operating
    batches are located at the last dimensions (trailing axes). For example, consider
    two second-order tensors in 3d-space as shown in Eq. :eq:`2d-tensors`.

    ..  math::
        :label: 2d-tensors

        \boldsymbol{T}_1 &= \begin{bmatrix} 
            0 & 3 & 6 \\
            1 & 4 & 7 \\
            2 & 5 & 8
        \end{bmatrix}
        
        \boldsymbol{T}_2 &= \begin{bmatrix} 
             9 & 12 & 15 \\
            10 & 13 & 16 \\
            11 & 14 & 17
        \end{bmatrix}
    
    This tensor is created with one elementwise-operating trailing axis.
    
    ..  plot::
        :context:
            
        >>> import tensortrax as tr
        >>> import numpy as np
        >>> 
        >>> T = tr.Tensor(np.arange(18).reshape(2, 3, 3).T, ntrax=1)
        >>> T
        <tensortrax tensor object>
          Shape / size of trailing axes: (2,) / 1
          Shape / size of tensor: (3, 3) / 9
        <BLANKLINE>
        Data (first batch):
        <BLANKLINE>
        array([[0, 3, 6],
               [1, 4, 7],
               [2, 5, 8]])
    
    A tensor, which is created without any dual data, is treated as constant and hence,
    all dual data arrays are initially filled with zeros.
    
    ..  tip::
        Broadcasting is used to enhance performance and to reduce memory consumption. If
        no dual data arrays are provided, the trailing axes of the dual arrays are
        compressed.
    
    ..  plot::
        :context:
            
        >>> T.x.shape
        (3, 3, 2)
        
        >>> T.δx.shape
        (3, 3, 1)
    
    If the gradient (or jacobian) w.r.t. this tensor should be recorded, the tensor must
    be initiated with the argument ``gradient=True``. This will add additional dual axes
    to the data arrays of the tensor. These axes are used for the gradient dimension.
    
    ..  plot::
        :context:
            
        >>> T.init(gradient=True)
        >>> T
        <tensortrax tensor object>
          Shape / size of trailing axes: (1, 1, 2) / 3
          Shape / size of tensor: (3, 3) / 9
        <BLANKLINE>
        Data (first batch):
        <BLANKLINE>
        array([[0, 3, 6],
               [1, 4, 7],
               [2, 5, 8]])
    
    Again, broadcasting is used, see the shapes of the tensor data arrays.
    
    ..  plot::
        :context:
            
        >>> T.x.shape
        (3, 3, 1, 1, 2)
        
        >>> T.δx.shape
        (3, 3, 3, 3, 1)
    
    If we look at the dual array, this is the (non-symmetric) fourth-order identity
    tensor, see Eq. :eq:`dual-identity`. This identity tensor has values of one located
    at the component to be differentiated and is filled with zeros otherwise.
    
    ..  math::
        :label: dual-identity
        
        \frac{\partial T_{ij}}{\partial T_{kl}} = \delta_{ik}\ \delta_{jl}
    
    ..  plot::
        :context:
        
        >>> T.δx[0, 1, ..., 0]
        array([[0., 1., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])
        
        >>> T.δx[1, 2, ..., 0]
        array([[0., 0., 0.],
               [0., 0., 1.],
               [0., 0., 0.]])
        
        >>> T.δx.reshape(9, 9)
        array([[1., 0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 1., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 1., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 1., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 1., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 1., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 1., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 1., 0.],
               [0., 0., 0., 0., 0., 0., 0., 0., 1.]])
    
    Now let's perform some math, e.g. the dot product as shown in Eq. :eq:`dot`. Along
    with the evaluation of the dot-product, the partial derivative of
    :math:`\boldsymbol{V}` w.r.t. :math:`\boldsymbol{T}` is recorded in the dual data
    array.
    
    ..  math::
        :label: dot
        
        \boldsymbol{V} &= \boldsymbol{T}^T\ \boldsymbol{T}
        
        \delta \boldsymbol{V} &=
            \delta \boldsymbol{T}^T\ \boldsymbol{T} + 
            \boldsymbol{T}^T\ \delta \boldsymbol{T}
    
    ..  plot::
        :context:
        
        >>> V = T.T @ T
        >>> V
        <tensortrax tensor object>
          Shape / size of trailing axes: (1, 1, 2) / 3
          Shape / size of tensor: (3, 3) / 9
        <BLANKLINE>
        Data (first batch):
        <BLANKLINE>
        array([[  5,  14,  23],
               [ 14,  50,  86],
               [ 23,  86, 149]])
        
        >>> V.δx[1, 2, ..., 0]
        array([[0., 6., 3.],
               [0., 7., 4.],
               [0., 8., 5.]])
    
    See Also
    --------
    tensortrax.function : Evaluate a function.
    tensortrax.gradient : Evaluate the gradient of a scalar-valued function.
    tensortrax.jacobian : Evaluate the jacobian of a function.
    tensortrax.hessian : Evaluate the hessian of a scalar-valued function.
    """

    def __init__(self, x, δx=None, Δx=None, Δδx=None, ntrax=0, ndual=0):
        self.x = np.asarray(x)

        self.ntrax = ntrax
        self.ndual = ndual

        self.shape = self.x.shape[: len(self.x.shape) - ntrax]
        self.trax = self.x.shape[len(self.x.shape) - ntrax :]
        self.size = np.prod(self.shape, dtype=int)

        self.δx = self._init_and_reshape(δx)
        self.Δx = self._init_and_reshape(Δx)
        self.Δδx = self._init_and_reshape(Δδx)

    def _add(self, ndual=1):
        "Add additional trailing axes for dual values."

        # reset the tensor
        self.reset()

        # add new dual-related axes to the shape of the data
        ones = np.ones(ndual, dtype=int)
        x = self.x.reshape(*self.shape, *ones, *self.trax)

        # re-init the tensor
        self.__init__(
            x,
            δx=self.δx,
            Δx=self.Δx,
            Δδx=self.Δδx,
            ntrax=self.ntrax + ndual,
            ndual=ndual,
        )

    def reset(self):
        "Reset the Tensor (set all dual values to zero)."

        # reset the shape of the data (delete axes related to dual data)
        x = self.x.reshape(*self.shape, *self.trax[self.ndual :])

        # re-init the tensor
        self.__init__(x, ntrax=self.ntrax - self.ndual)

    def copy(self):
        "Copy the Tensor."

        return deepcopy(self)

    def init(self, gradient=False, hessian=False, sym=False, δx=None, Δx=None):
        """Re-Initialize tensor with dual values to keep track of the
        hessian and/or the gradient."""

        if gradient and not hessian:
            # add additional element-wise acting axes for dual values
            self._add(ndual=len(self.shape))

            # create the dual values
            if δx is None or isinstance(δx, bool):
                shape = (*self.shape, *self.shape)
                if len(shape) == 0:
                    shape = (1,)
                if δx is False:
                    δx = np.zeros(self.size**2).reshape(shape)
                else:
                    δx = np.eye(self.size).reshape(shape)

            else:
                δx = δx.reshape(*self.shape, *self.trax)

            Δx = δx.copy()

        elif hessian:
            # add additional trailing axes for dual values
            self._add(ndual=2 * len(self.shape))

            # create the dual values
            ones = np.ones(len(self.shape), dtype=int)

            if δx is None or isinstance(δx, bool):
                shape = (*self.shape, *self.shape, *ones)
                if len(shape) == 0:
                    shape = (1,)
                if δx is False:
                    δx = np.zeros(self.size**2).reshape(shape)
                else:
                    δx = np.eye(self.size).reshape(shape)
            else:
                δx = δx.reshape(*self.shape, *self.trax)

            if Δx is None or isinstance(Δx, bool):
                shape = (*self.shape, *ones, *self.shape)
                if len(shape) == 0:
                    shape = (1,)
                if Δx is False:
                    Δx = np.zeros(self.size**2).reshape(shape)
                else:
                    Δx = np.eye(self.size).reshape(shape)
            else:
                Δx = Δx.reshape(*self.shape, *self.trax)

        if gradient or hessian:
            if sym:
                idx_off_diag = {1: None, 3: [1], 6: [1, 2, 4]}[self.size]
                δx[idx_off_diag] /= 2
                Δx[idx_off_diag] /= 2

            # re-init the tensor
            self.__init__(self.x, δx=δx, Δx=Δx, ntrax=self.ntrax, ndual=self.ndual)

    def _init_and_reshape(self, value):
        if value is None:
            value = np.zeros(self.shape)
        else:
            value = np.asarray(value)
        if len(value.shape) != len(self.x.shape):
            newshape = (
                *value.shape,
                *np.ones(len(self.x.shape) - len(value.shape), dtype=int),
            )
            value = value.reshape(newshape)
        return value

    def __neg__(self):
        return Tensor(
            x=-self.x, δx=-self.δx, Δx=-self.Δx, Δδx=-self.Δδx, ntrax=self.ntrax
        )

    def __add__(self, B):
        A = self
        if isinstance(B, Tensor):
            x = f(A) + f(B)
            δx = δ(A) + δ(B)
            Δx = Δ(A) + Δ(B)
            Δδx = Δδ(A) + Δδ(B)
            ntrax = min(A.ntrax, B.ntrax)
        else:
            x = f(A) + B
            δx = δ(A)
            Δx = Δ(A)
            Δδx = Δδ(A)
            ntrax = A.ntrax
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)

    def __sub__(self, B):
        A = self
        if isinstance(B, Tensor):
            x = f(A) - f(B)
            δx = δ(A) - δ(B)
            Δx = Δ(A) - Δ(B)
            Δδx = Δδ(A) - Δδ(B)
            ntrax = min(A.ntrax, B.ntrax)
        else:
            x = f(A) - B
            δx = δ(A)
            Δx = Δ(A)
            Δδx = Δδ(A)
            ntrax = A.ntrax
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)

    def __rsub__(self, B):
        return -self.__sub__(B)

    def __mul__(self, B):
        A = self
        if isinstance(B, Tensor):
            x = f(A) * f(B)
            δx = δ(A) * f(B) + f(A) * δ(B)
            Δx = Δ(A) * f(B) + f(A) * Δ(B)
            Δδx = Δ(A) * δ(B) + δ(A) * Δ(B) + Δδ(A) * f(B) + f(A) * Δδ(B)
            ntrax = min(A.ntrax, B.ntrax)
        else:
            x = f(A) * B
            δx = δ(A) * B
            Δx = Δ(A) * B
            Δδx = Δδ(A) * B
            ntrax = A.ntrax
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)

    def __truediv__(self, B):
        A = self
        return A * B**-1

    def __rtruediv__(self, B):
        A = self
        return B * A**-1

    def __pow__(self, p):
        A = self
        x = f(A) ** p
        δx = p * f(A) ** (p - 1) * δ(A)
        Δx = p * f(A) ** (p - 1) * Δ(A)
        Δδx = p * f(A) ** (p - 1) * Δδ(A) + p * (p - 1) * f(A) ** (p - 2) * δ(A) * Δ(A)
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=A.ntrax)

    def __gt__(self, B):
        A = self
        return f(A) > (f(B) if isinstance(B, Tensor) else B)

    def __lt__(self, B):
        A = self
        return f(A) < (f(B) if isinstance(B, Tensor) else B)

    def __ge__(self, B):
        A = self
        return f(A) >= (f(B) if isinstance(B, Tensor) else B)

    def __le__(self, B):
        A = self
        return f(A) <= (f(B) if isinstance(B, Tensor) else B)

    def __eq__(self, B):
        A = self
        return f(A) == (f(B) if isinstance(B, Tensor) else B)

    def __matmul__(self, B):
        return matmul(self, B)

    def __rmatmul__(self, B):
        return matmul(B, self)

    def __getitem__(self, key):
        x = f(self)[key]
        Δx = Δ(self)[key]
        δx = δ(self)[key]
        Δδx = Δδ(self)[key]
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=self.ntrax)

    def __setitem__(self, key, value):
        if isinstance(value, Tensor):
            shape = (*self.shape, *self.trax)
            self.x[key] = f(value)
            if self.δx[key].shape != δ(value).shape:
                self.δx = broadcast_to(self.δx, shape).copy()
            if self.Δx[key].shape != Δ(value).shape:
                self.Δx = broadcast_to(self.Δx, shape).copy()
            if self.Δδx[key].shape != Δδ(value).shape:
                self.Δδx = broadcast_to(self.Δδx, shape).copy()
            self.δx[key] = δ(value)
            self.Δx[key] = Δ(value)
            self.Δδx[key] = Δδ(value)
        else:
            self.x[key] = value
            self.δx[key].fill(0)
            self.Δx[key].fill(0)
            self.Δδx[key].fill(0)

    def __repr__(self):
        header = "<tensortrax tensor object>"
        metadata = [
            f"  Shape / size of trailing axes: {self.trax} / {self.ntrax}",
            f"  Shape / size of tensor: {self.shape} / {self.size}",
            "",
            "Data (first batch):",
        ]
        first_batch = (0,) * self.ntrax
        data = self.x[(..., *first_batch)].__repr__()
        return "\n".join([header, *metadata, "", data])

    @property
    def T(self):
        return transpose(self)

    def ravel(self, order="C"):
        "Return a contiguous flattened array."
        return ravel(self, order=order)

    def reshape(self, *shape, order="C"):
        "Gives a new shape to an array without changing its data."
        return reshape(self, shape, order=order)

    def squeeze(self, axis=None):
        "Remove axes of length one."
        return squeeze(self, axis=axis)

    def dual2real(self, like):
        return dual_to_real(self, like=like)

    def dual_to_real(self, like):
        return dual_to_real(self, like=like)

    def real_to_dual(self, x, mul=None):
        return real_to_dual(self, x=x, mul=mul)

    def astype(self, dtype):
        return Tensor(
            x=self.x.astype(dtype),
            δx=self.δx.astype(dtype),
            Δx=self.Δx.astype(dtype),
            Δδx=self.Δδx.astype(dtype),
            ntrax=self.ntrax,
        )

    __radd__ = __add__
    __rmul__ = __mul__
    __array_ufunc__ = None


def broadcast_to(A, shape):
    "Broadcast Array or Tensor to a new shape."

    def _broadcast_to(a):
        return np.broadcast_to(a, shape=shape)

    if isinstance(A, Tensor):
        return Tensor(
            x=_broadcast_to(f(A)),
            δx=_broadcast_to(δ(A)),
            Δx=_broadcast_to(Δ(A)),
            Δδx=_broadcast_to(Δδ(A)),
            ntrax=A.ntrax,
        )
    else:
        return _broadcast_to(A)


def dual_to_real(A, like=None):
    """Return a new Tensor with old-dual data as new-real values,
    with `ntrax` derived by `like`."""

    ndual = like.ndual - len(like.shape)
    ntrax = A.ntrax - ndual

    return Tensor(x=A.δx, δx=A.Δδx, Δx=A.Δδx, ndual=min(ndual, ntrax), ntrax=ntrax)


def real_to_dual(A, x, mul=None):
    r"""Return a new Tensor with old-real data as new-dual values.

    ..  math::

        \delta W(x) &= A(x)\ \delta x

        \Delta \delta W(x) &= \Delta A(x)\ \delta x + A(x)\ \Delta \delta x

    """

    if mul is None:

        def mul(A, B):
            return A * B

    return Tensor(
        x=mul(f(A), f(x)) * np.nan,
        δx=mul(f(A), δ(x)),
        Δx=mul(f(A), Δ(x)),
        Δδx=mul(δ(A), Δ(x)) + mul(f(A), Δδ(x)),
        ntrax=A.ntrax,
    )


def ravel(A, order="C"):
    "Return a contiguous flattened array."
    if isinstance(A, Tensor):
        δtrax = δ(A).shape[len(A.shape) :]
        Δtrax = Δ(A).shape[len(A.shape) :]
        Δδtrax = Δδ(A).shape[len(A.shape) :]
        return Tensor(
            x=f(A).reshape(A.size, *A.trax, order=order),
            δx=δ(A).reshape(A.size, *δtrax, order=order),
            Δx=Δ(A).reshape(A.size, *Δtrax, order=order),
            Δδx=Δδ(A).reshape(A.size, *Δδtrax, order=order),
            ntrax=A.ntrax,
        )
    else:
        return np.ravel(A, order=order)


def squeeze(A, axis=None):
    "Remove axes of length one."
    if isinstance(A, Tensor):
        if axis is None:
            if 1 in A.shape:
                axis = tuple(np.arange(len(A.shape))[np.array(A.shape) == 1])
            else:
                axis = ()
        return Tensor(
            x=np.squeeze(f(A), axis=axis),
            δx=np.squeeze(δ(A), axis=axis),
            Δx=np.squeeze(Δ(A), axis=axis),
            Δδx=np.squeeze(Δδ(A), axis=axis),
            ntrax=A.ntrax,
        )
    else:
        return np.squeeze(A, axis=axis)


def reshape(A, shape, order="C"):
    "Gives a new shape to an array without changing its data."
    if isinstance(A, Tensor):
        δtrax = δ(A).shape[len(A.shape) :]
        Δtrax = Δ(A).shape[len(A.shape) :]
        Δδtrax = Δδ(A).shape[len(A.shape) :]
        return Tensor(
            x=f(A).reshape(*shape, *A.trax, order=order),
            δx=δ(A).reshape(*shape, *δtrax, order=order),
            Δx=Δ(A).reshape(*shape, *Δtrax, order=order),
            Δδx=Δδ(A).reshape(*shape, *Δδtrax, order=order),
            ntrax=A.ntrax,
        )
    else:
        return np.reshape(A, shape, order=order)


def einsum4(subscripts, *operands):
    "Einsum with four operands."
    A, B, C, D = operands

    def _einsum(*operands):
        return np.einsum(subscripts, *operands)

    if (
        isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(f(A), f(B), f(C), f(D))
        δx = (
            _einsum(δ(A), f(B), f(C), f(D))
            + _einsum(f(A), δ(B), f(C), f(D))
            + _einsum(f(A), f(B), δ(C), f(D))
            + _einsum(f(A), f(B), f(C), δ(D))
        )
        Δx = (
            _einsum(Δ(A), f(B), f(C), f(D))
            + _einsum(f(A), Δ(B), f(C), f(D))
            + _einsum(f(A), f(B), Δ(C), f(D))
            + _einsum(f(A), f(B), f(C), Δ(D))
        )
        Δδx = (
            _einsum(Δδ(A), f(B), f(C), f(D))
            + _einsum(f(A), Δδ(B), f(C), f(D))
            + _einsum(f(A), f(B), Δδ(C), f(D))
            + _einsum(f(A), f(B), f(C), Δδ(D))
            + _einsum(δ(A), Δ(B), f(C), f(D))
            + _einsum(Δ(A), δ(B), f(C), f(D))
            + _einsum(δ(A), f(B), Δ(C), f(D))
            + _einsum(Δ(A), f(B), δ(C), f(D))
            + _einsum(δ(A), f(B), f(C), Δ(D))
            + _einsum(Δ(A), f(B), f(C), δ(D))
            + _einsum(f(A), δ(B), Δ(C), f(D))
            + _einsum(f(A), Δ(B), δ(C), f(D))
            + _einsum(f(A), δ(B), f(C), Δ(D))
            + _einsum(f(A), Δ(B), f(C), δ(D))
            + _einsum(f(A), f(B), δ(C), Δ(D))
            + _einsum(f(A), f(B), Δ(C), δ(D))
        )
        ntrax = min(A.ntrax, B.ntrax, C.ntrax, D.ntrax)
    elif (
        isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(f(A), B, C, D)
        δx = _einsum(δ(A), B, C, D)
        Δx = _einsum(Δ(A), B, C, D)
        Δδx = _einsum(Δδ(A), B, C, D)
        ntrax = A.ntrax
    elif (
        not isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(A, f(B), C, D)
        δx = _einsum(A, δ(B), C, D)
        Δx = _einsum(A, Δ(B), C, D)
        Δδx = _einsum(A, Δδ(B), C, D)
        ntrax = B.ntrax
    elif (
        not isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(A, B, f(C), D)
        δx = _einsum(A, B, δ(C), D)
        Δx = _einsum(A, B, Δ(C), D)
        Δδx = _einsum(A, B, Δδ(C), D)
        ntrax = C.ntrax
    elif (
        not isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(A, B, C, f(D))
        δx = _einsum(A, B, C, δ(D))
        Δx = _einsum(A, B, C, Δ(D))
        Δδx = _einsum(A, B, C, Δδ(D))
        ntrax = D.ntrax
    elif (
        isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(f(A), f(B), C, D)
        δx = _einsum(δ(A), f(B), C, D) + _einsum(f(A), δ(B), C, D)
        Δx = _einsum(Δ(A), f(B), C, D) + _einsum(f(A), Δ(B), C, D)
        Δδx = (
            _einsum(Δδ(A), f(B), C, D)
            + _einsum(f(A), Δδ(B), C, D)
            + _einsum(δ(A), Δ(B), C, D)
            + _einsum(Δ(A), δ(B), C, D)
        )
        ntrax = min(A.ntrax, B.ntrax)
    elif (
        isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(f(A), B, f(C), D)
        δx = _einsum(δ(A), B, f(C), D) + _einsum(f(A), B, δ(C), D)
        Δx = _einsum(Δ(A), B, f(C), D) + _einsum(f(A), B, Δ(C), D)
        Δδx = (
            _einsum(Δδ(A), B, f(C), D)
            + _einsum(f(A), B, Δδ(C), D)
            + _einsum(δ(A), B, Δ(C), D)
            + _einsum(Δ(A), B, δ(C), D)
        )
        ntrax = min(A.ntrax, C.ntrax)
    elif (
        isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(f(A), B, C, f(D))
        δx = _einsum(δ(A), B, C, f(D)) + _einsum(f(A), B, C, δ(D))
        Δx = _einsum(Δ(A), B, C, f(D)) + _einsum(f(A), B, C, Δ(D))
        Δδx = (
            _einsum(Δδ(A), B, C, f(D))
            + _einsum(f(A), B, C, Δδ(D))
            + _einsum(δ(A), B, C, Δ(D))
            + _einsum(Δ(A), B, C, δ(D))
        )
        ntrax = min(A.ntrax, D.ntrax)
    elif (
        not isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(A, f(B), f(C), D)
        δx = _einsum(A, δ(B), f(C), D) + _einsum(A, f(B), δ(C), D)
        Δx = _einsum(A, Δ(B), f(C), D) + _einsum(A, f(B), Δ(C), D)
        Δδx = (
            _einsum(A, Δδ(B), f(C), D)
            + _einsum(A, f(B), Δδ(C), D)
            + _einsum(A, δ(B), Δ(C), D)
            + _einsum(A, Δ(B), δ(C), D)
        )
        ntrax = min(B.ntrax, C.ntrax)
    elif (
        not isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(A, f(B), C, f(D))
        δx = _einsum(A, δ(B), C, f(D)) + _einsum(A, f(B), C, δ(D))
        Δx = _einsum(A, Δ(B), C, f(D)) + _einsum(A, f(B), C, Δ(D))
        Δδx = (
            _einsum(A, Δδ(B), C, f(D))
            + _einsum(A, f(B), C, Δδ(D))
            + _einsum(A, δ(B), C, Δ(D))
            + _einsum(A, Δ(B), C, δ(D))
        )
        ntrax = min(B.ntrax, D.ntrax)
    elif (
        not isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(A, B, f(C), f(D))
        δx = _einsum(A, B, δ(C), f(D)) + _einsum(A, B, f(C), δ(D))
        Δx = _einsum(A, B, Δ(C), f(D)) + _einsum(A, B, f(C), Δ(D))
        Δδx = (
            _einsum(A, B, Δδ(C), f(D))
            + _einsum(A, B, f(C), Δδ(D))
            + _einsum(A, B, δ(C), Δ(D))
            + _einsum(A, B, Δ(C), δ(D))
        )
        ntrax = min(C.ntrax, D.ntrax)
    elif (
        isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and not isinstance(D, Tensor)
    ):
        x = _einsum(f(A), f(B), f(C), D)
        δx = (
            _einsum(δ(A), f(B), f(C), D)
            + _einsum(f(A), δ(B), f(C), D)
            + _einsum(f(A), f(B), δ(C), D)
        )
        Δx = (
            _einsum(Δ(A), f(B), f(C), D)
            + _einsum(f(A), Δ(B), f(C), D)
            + _einsum(f(A), f(B), Δ(C), D)
        )
        Δδx = (
            _einsum(Δδ(A), f(B), f(C), D)
            + _einsum(f(A), Δδ(B), f(C), D)
            + _einsum(f(A), f(B), Δδ(C), D)
            + _einsum(δ(A), Δ(B), f(C), D)
            + _einsum(Δ(A), δ(B), f(C), D)
            + _einsum(δ(A), f(B), Δ(C), D)
            + _einsum(Δ(A), f(B), δ(C), D)
            + _einsum(f(A), δ(B), Δ(C), D)
            + _einsum(f(A), Δ(B), δ(C), D)
        )
        ntrax = min(A.ntrax, B.ntrax, C.ntrax)
    elif (
        isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and not isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(f(A), f(B), C, f(D))
        δx = (
            _einsum(δ(A), f(B), C, f(D))
            + _einsum(f(A), δ(B), C, f(D))
            + _einsum(f(A), f(B), C, δ(D))
        )
        Δx = (
            _einsum(Δ(A), f(B), C, f(D))
            + _einsum(f(A), Δ(B), C, f(D))
            + _einsum(f(A), f(B), C, Δ(D))
        )
        Δδx = (
            _einsum(Δδ(A), f(B), C, f(D))
            + _einsum(f(A), Δδ(B), C, f(D))
            + _einsum(f(A), f(B), C, Δδ(D))
            + _einsum(δ(A), Δ(B), C, f(D))
            + _einsum(Δ(A), δ(B), C, f(D))
            + _einsum(δ(A), f(B), C, Δ(D))
            + _einsum(Δ(A), f(B), C, δ(D))
            + _einsum(f(A), δ(B), C, Δ(D))
            + _einsum(f(A), Δ(B), C, δ(D))
        )
        ntrax = min(A.ntrax, B.ntrax, D.ntrax)
    elif (
        isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(f(A), B, f(C), f(D))
        δx = (
            _einsum(δ(A), B, f(C), f(D))
            + _einsum(f(A), B, δ(C), f(D))
            + _einsum(f(A), B, f(C), δ(D))
        )
        Δx = (
            _einsum(Δ(A), B, f(C), f(D))
            + _einsum(f(A), B, Δ(C), f(D))
            + _einsum(f(A), B, f(C), Δ(D))
        )
        Δδx = (
            _einsum(Δδ(A), B, f(C), f(D))
            + _einsum(f(A), B, Δδ(C), f(D))
            + _einsum(f(A), B, f(C), Δδ(D))
            + _einsum(δ(A), B, Δ(C), f(D))
            + _einsum(Δ(A), B, δ(C), f(D))
            + _einsum(δ(A), B, f(C), Δ(D))
            + _einsum(Δ(A), B, f(C), δ(D))
            + _einsum(f(A), B, δ(C), Δ(D))
            + _einsum(f(A), B, Δ(C), δ(D))
        )
        ntrax = min(A.ntrax, C.ntrax, D.ntrax)
    elif (
        not isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and isinstance(C, Tensor)
        and isinstance(D, Tensor)
    ):
        x = _einsum(A, f(B), f(C), f(D))
        δx = (
            _einsum(A, δ(B), f(C), f(D))
            + _einsum(A, f(B), δ(C), f(D))
            + _einsum(A, f(B), f(C), δ(D))
        )
        Δx = (
            _einsum(A, Δ(B), f(C), f(D))
            + _einsum(A, f(B), Δ(C), f(D))
            + _einsum(A, f(B), f(C), Δ(D))
        )
        Δδx = (
            _einsum(A, Δδ(B), f(C), f(D))
            + _einsum(A, f(B), Δδ(C), f(D))
            + _einsum(A, f(B), f(C), Δδ(D))
            + _einsum(A, δ(B), Δ(C), f(D))
            + _einsum(A, Δ(B), δ(C), f(D))
            + _einsum(A, δ(B), f(C), Δ(D))
            + _einsum(A, Δ(B), f(C), δ(D))
            + _einsum(A, f(B), δ(C), Δ(D))
            + _einsum(A, f(B), Δ(C), δ(D))
        )
        ntrax = min(B.ntrax, C.ntrax, D.ntrax)
    else:
        return _einsum(*operands)

    return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)


def einsum3(subscripts, *operands):
    "Einsum with three operands."
    A, B, C = operands

    def _einsum(*operands):
        return np.einsum(subscripts, *operands)

    if isinstance(A, Tensor) and isinstance(B, Tensor) and isinstance(C, Tensor):
        x = _einsum(f(A), f(B), f(C))
        δx = (
            _einsum(δ(A), f(B), f(C))
            + _einsum(f(A), δ(B), f(C))
            + _einsum(f(A), f(B), δ(C))
        )
        Δx = (
            _einsum(Δ(A), f(B), f(C))
            + _einsum(f(A), Δ(B), f(C))
            + _einsum(f(A), f(B), Δ(C))
        )
        Δδx = (
            _einsum(Δδ(A), f(B), f(C))
            + _einsum(f(A), Δδ(B), f(C))
            + _einsum(f(A), f(B), Δδ(C))
            + _einsum(δ(A), Δ(B), f(C))
            + _einsum(Δ(A), δ(B), f(C))
            + _einsum(δ(A), f(B), Δ(C))
            + _einsum(Δ(A), f(B), δ(C))
            + _einsum(f(A), δ(B), Δ(C))
            + _einsum(f(A), Δ(B), δ(C))
        )
        ntrax = min(A.ntrax, B.ntrax, C.ntrax)
    elif (
        isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and not isinstance(C, Tensor)
    ):
        x = _einsum(f(A), B, C)
        δx = _einsum(δ(A), B, C)
        Δx = _einsum(Δ(A), B, C)
        Δδx = _einsum(Δδ(A), B, C)
        ntrax = A.ntrax
    elif (
        not isinstance(A, Tensor)
        and isinstance(B, Tensor)
        and not isinstance(C, Tensor)
    ):
        x = _einsum(A, f(B), C)
        δx = _einsum(A, δ(B), C)
        Δx = _einsum(A, Δ(B), C)
        Δδx = _einsum(A, Δδ(B), C)
        ntrax = B.ntrax
    elif (
        not isinstance(A, Tensor)
        and not isinstance(B, Tensor)
        and isinstance(C, Tensor)
    ):
        x = _einsum(A, B, f(C))
        δx = _einsum(A, B, δ(C))
        Δx = _einsum(A, B, Δ(C))
        Δδx = _einsum(A, B, Δδ(C))
        ntrax = C.ntrax
    elif isinstance(A, Tensor) and isinstance(B, Tensor) and not isinstance(C, Tensor):
        x = _einsum(f(A), f(B), C)
        δx = _einsum(δ(A), f(B), C) + _einsum(f(A), δ(B), C)
        Δx = _einsum(Δ(A), f(B), C) + _einsum(f(A), Δ(B), C)
        Δδx = (
            _einsum(Δδ(A), f(B), C)
            + _einsum(f(A), Δδ(B), C)
            + _einsum(δ(A), Δ(B), C)
            + _einsum(Δ(A), δ(B), C)
        )
        ntrax = min(A.ntrax, B.ntrax)
    elif isinstance(A, Tensor) and not isinstance(B, Tensor) and isinstance(C, Tensor):
        x = _einsum(f(A), B, f(C))
        δx = _einsum(δ(A), B, f(C)) + _einsum(f(A), B, δ(C))
        Δx = _einsum(Δ(A), B, f(C)) + _einsum(f(A), B, Δ(C))
        Δδx = (
            _einsum(Δδ(A), B, f(C))
            + _einsum(f(A), B, Δδ(C))
            + _einsum(δ(A), B, Δ(C))
            + _einsum(Δ(A), B, δ(C))
        )
        ntrax = min(A.ntrax, C.ntrax)
    elif not isinstance(A, Tensor) and isinstance(B, Tensor) and isinstance(C, Tensor):
        x = _einsum(A, f(B), f(C))
        δx = _einsum(A, δ(B), f(C)) + _einsum(A, f(B), δ(C))
        Δx = _einsum(A, Δ(B), f(C)) + _einsum(A, f(B), Δ(C))
        Δδx = (
            _einsum(A, Δδ(B), f(C))
            + _einsum(A, f(B), Δδ(C))
            + _einsum(A, δ(B), Δ(C))
            + _einsum(A, Δ(B), δ(C))
        )
        ntrax = min(B.ntrax, C.ntrax)
    else:
        return _einsum(*operands)

    return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)


def einsum2(subscripts, *operands):
    "Einsum with two operands."
    A, B = operands

    def _einsum(*operands):
        return np.einsum(subscripts, *operands)

    if isinstance(A, Tensor) and isinstance(B, Tensor):
        x = _einsum(f(A), f(B))
        δx = _einsum(δ(A), f(B)) + _einsum(f(A), δ(B))
        Δx = _einsum(Δ(A), f(B)) + _einsum(f(A), Δ(B))
        Δδx = (
            _einsum(Δδ(A), f(B))
            + _einsum(f(A), Δδ(B))
            + _einsum(δ(A), Δ(B))
            + _einsum(Δ(A), δ(B))
        )
        ntrax = min(A.ntrax, B.ntrax)
    elif isinstance(A, Tensor) and not isinstance(B, Tensor):
        x = _einsum(f(A), B)
        δx = _einsum(δ(A), B)
        Δx = _einsum(Δ(A), B)
        Δδx = _einsum(Δδ(A), B)
        ntrax = A.ntrax
    elif not isinstance(A, Tensor) and isinstance(B, Tensor):
        x = _einsum(A, f(B))
        δx = _einsum(A, δ(B))
        Δx = _einsum(A, Δ(B))
        Δδx = _einsum(A, Δδ(B))
        ntrax = B.ntrax
    else:
        return _einsum(*operands)

    return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=ntrax)


def einsum1(subscripts, *operands):
    "Einsum with one operand."
    A = operands[0]

    def _einsum(*operands):
        return np.einsum(subscripts, *operands)

    if isinstance(A, Tensor):
        x = _einsum(f(A))
        δx = _einsum(δ(A))
        Δx = _einsum(Δ(A))
        Δδx = _einsum(Δδ(A))
        return Tensor(x=x, δx=δx, Δx=Δx, Δδx=Δδx, ntrax=A.ntrax)
    else:
        return _einsum(*operands)


def einsum(subscripts, *operands):
    "Einsum limited to one, two, three or four operands."
    if len(operands) == 1:
        return einsum1(subscripts, *operands)
    elif len(operands) == 2:
        return einsum2(subscripts, *operands)
    elif len(operands) == 3:
        return einsum3(subscripts, *operands)
    elif len(operands) == 4:
        return einsum4(subscripts, *operands)
    else:
        raise NotImplementedError("More than four operands are not supported.")


def transpose(A):
    "Returns an array with axes transposed."
    ij = "abcdefghijklmnopqrstuvwxyz"[: len(A.shape)]
    ji = ij[::-1]
    return einsum(f"{ij}...->{ji}...", A)


def matmul(A, B):
    "Matrix product of two arrays."
    ik = "ik"[2 - len(A.shape) :]
    kj = "kj"[: len(B.shape)]
    ij = (ik + kj).replace("k", "")
    return einsum(f"{ik}...,{kj}...->{ij}...", A, B)



"""
tensorTRAX: Math on (Hyper-Dual) Tensors with Trailing Axes.
"""

import numpy as np

from ..._tensor import Tensor, Δ, Δδ, einsum, f, matmul, δ
from .._math_tensor import exp, sqrt, sum, transpose
from ..special._special_tensor import ddot
from . import _linalg_array as linalg

dot = matmul


def det(A):
    "Determinant of a 2x2 or 3x3 Tensor."
    if isinstance(A, Tensor):
        x = linalg.det(f(A))
        B = transpose(linalg.inv(f(A)))
        δx = x * ddot(B, δ(A))
        Δx = x * ddot(B, Δ(A))

        ΔB = -matmul(matmul(B, transpose(Δ(A))), B)
        Δδx = Δx * δx / x + x * ddot(ΔB, δ(A)) + x * ddot(B, Δδ(A))
        return Tensor(
            x=x,
            δx=δx,
            Δx=Δx,
            Δδx=Δδx,
            ntrax=A.ntrax,
        )
    else:
        return linalg.det(A)


def inv(A, inverse=linalg.inv):
    "Inverse of a 2x2 or 3x3 Tensor."
    if isinstance(A, Tensor):
        x = inverse(f(A))
        invA = inverse(f(A))
        δx = -matmul(matmul(invA, δ(A)), invA)
        Δx = -matmul(matmul(invA, Δ(A)), invA)
        Δδx = -(
            matmul(matmul(Δx, δ(A)), invA)
            + matmul(matmul(invA, δ(A)), Δx)
            + matmul(matmul(invA, Δδ(A)), invA)
        )

        return Tensor(
            x=x,
            δx=δx,
            Δx=Δx,
            Δδx=Δδx,
            ntrax=A.ntrax,
        )
    else:
        return linalg.inv(A)


def pinv(A):
    "Pseudo-Inverse of a 2x2 or 3x3 Tensor."

    return inv(A, inverse=linalg.pinv)


def eigvalsh(A, eps=np.sqrt(np.finfo(float).eps)):
    "Eigenvalues of a symmetric Tensor."

    if isinstance(A, Tensor):
        A[0, 0] += eps
        A[1, 1] -= eps

        λ, N = [x.T for x in np.linalg.eigh(f(A).T)]
        M = einsum("ai...,aj...->aij...", N, N)

        dim = len(λ)

        δλ = einsum("aij...,ij...->a...", M, δ(A))
        Δλ = einsum("aij...,ij...->a...", M, Δ(A))

        # alpha = [0, 1, 2]
        # beta = [(1, 2), (2, 0), (0, 1)]

        alpha = np.arange(dim)
        beta = [
            np.concatenate([np.arange(a + 1, dim), np.arange(a)])
            for a in np.arange(dim)
        ]

        δN = []
        for α in alpha:
            δNα = []
            for β in beta[α]:
                Mαβ = einsum("i...,j...->ij...", N[α], N[β])
                δAαβ = einsum("ij...,ij...->...", Mαβ, δ(A))
                λαβ = λ[α] - λ[β]
                δNα.append(1 / λαβ * N[β] * δAαβ)
            δN.append(sum(δNα, axis=0))

        δM = einsum("ai...,aj...->aij...", δN, N) + einsum("ai...,aj...->aij...", N, δN)
        Δδλ = einsum("aij...,ij...->a...", δM, Δ(A)) + einsum(
            "aij...,ij...->a...", M, Δδ(A)
        )

        return Tensor(
            x=λ,
            δx=δλ,
            Δx=Δλ,
            Δδx=Δδλ,
            ntrax=A.ntrax,
        )

    else:
        return np.linalg.eigvalsh(A.T).T


def eigh(A, eps=np.sqrt(np.finfo(float).eps)):
    "Eigenvalues and -bases of a symmetric Tensor."

    if isinstance(A, Tensor):
        A[0, 0] += eps
        A[1, 1] -= eps

        λ, N = [x.T for x in np.linalg.eigh(f(A).T)]
        M = einsum("ai...,aj...->aij...", N, N)

        dim = len(λ)

        δλ = einsum("aij...,ij...->a...", M, δ(A))
        Δλ = einsum("aij...,ij...->a...", M, Δ(A))

        # alpha = [0, 1, 2]
        # beta = [(1, 2), (2, 0), (0, 1)]
        alpha = np.arange(dim)
        beta = [
            np.concatenate([np.arange(a + 1, dim), np.arange(a)])
            for a in np.arange(dim)
        ]

        δN = []
        ΔN = []
        for α in alpha:
            δNα = []
            ΔNα = []
            for β in beta[α]:
                Mαβ = einsum("i...,j...->ij...", N[α], N[β])
                δAαβ = einsum("ij...,ij...->...", Mαβ, δ(A))
                ΔAαβ = einsum("ij...,ij...->...", Mαβ, Δ(A))
                λαβ = λ[α] - λ[β]
                δNα.append(1 / λαβ * N[β] * δAαβ)
                ΔNα.append(1 / λαβ * N[β] * ΔAαβ)
            δN.append(sum(δNα, axis=0))
            ΔN.append(sum(ΔNα, axis=0))

        ΔδN = []
        for α in alpha:
            ΔδNα = []
            for β in beta[α]:
                Mαβ = einsum("i...,j...->ij...", N[α], N[β])
                δAαβ = einsum("ij...,ij...->...", Mαβ, δ(A))
                ΔδAαβ = einsum("ij...,ij...->...", Mαβ, Δδ(A))
                λαβ = λ[α] - λ[β]
                Δλαβ = Δλ[α] - Δλ[β]
                ΔδNα.append(
                    -(λαβ**-2) * Δλαβ * N[β] * δAαβ
                    + 1 / λαβ * ΔN[β] * δAαβ
                    + 1 / λαβ * N[β] * ΔδAαβ
                )
            ΔδN.append(sum(ΔδNα, axis=0))

        δM = einsum("ai...,aj...->aij...", δN, N) + einsum("ai...,aj...->aij...", N, δN)
        ΔM = einsum("ai...,aj...->aij...", ΔN, N) + einsum("ai...,aj...->aij...", N, ΔN)
        Δδλ = einsum("aij...,ij...->a...", δM, Δ(A)) + einsum(
            "aij...,ij...->a...", M, Δδ(A)
        )

        ΔδM = (
            einsum("ai...,aj...->aij...", δN, ΔN)
            + einsum("ai...,aj...->aij...", ΔN, δN)
            + einsum("ai...,aj...->aij...", ΔδN, N)
            + einsum("ai...,aj...->aij...", N, ΔδN)
        )

        return (
            Tensor(
                x=λ,
                δx=δλ,
                Δx=Δλ,
                Δδx=Δδλ,
                ntrax=A.ntrax,
            ),
            Tensor(
                x=M,
                δx=δM,
                Δx=ΔM,
                Δδx=ΔδM,
                ntrax=A.ntrax,
            ),
        )

    else:
        λ, N = [x.T for x in np.linalg.eigh(A.T)]
        M = einsum("ai...,aj...->aij...", N, N)
        return λ, M


def expm(A):
    "Compute the matrix exponential of a symmetric array."
    λ, M = eigh(A)
    return einsum("a...,aij...->ij...", exp(λ), M)


def sqrtm(A):
    "Compute the matrix square root of a symmetric array."
    λ, M = eigh(A)
    return einsum("a...,aij...->ij...", sqrt(λ), M)



from ._linalg_array import det as _det
from ._linalg_array import inv as _inv
from ._linalg_array import pinv as _pinv
from ._linalg_tensor import det, eigh, eigvalsh, expm, inv, pinv, sqrtm

__all__ = [
    "_det",
    "_inv",
    "_pinv",
    "det",
    "eigh",
    "eigvalsh",
    "expm",
    "inv",
    "pinv",
    "sqrtm",
]
